# coding=utf-8
"""
The `SV` module is used to define classes and methods specific to
a SystemVision analog flow

.. highlight:: xml

**Input Files**:
----------------------

**Design File**:
This is the main design file used to import geometry.
Most shapes and vias will come from this file.
Examples: Allegro (.mcm, .brd), ODB++ (.zip), DXF (.dxf).
Set using the 'designFile' attribute.
See `DesignFlow.importDesign`.


**Technology File** (Optional):
This file is used to fix layer/materials information
that may be invalid from the design file.
This can be exported/imported from/to
the HyperLynx Advanced Solver tools.
Example: 'technology.tech'.
Set using the 'techFile' attribute.
See `DesignFlow.postImportOperations`.


**Boolean Geometry File(s)** (Optional):
The BOOL format file is a text-based geometry file used to
add Bond Wires and Pins that are not present in the design file.
This can be imported into the HyperLynx Advanced Solver tools. Format::

    UNITS micron
    NETS Yes
    # Comment:
    BONDWIRE bondWireProfileName netName xStart yStart xEnd yEnd
    PIN "name"
    COMPONENT_NAME "name"
    RECTANGLE layerName netName ADD width height xCenter yCenter angle

Example: 'boolfile.bool'.
Set using the 'boolFiles' attribute.
See `DesignFlow._readBoolFiles`.
Note: Overlapping Bond Wires may cause issues in the flow due to overlapping Nets.


**Lead Finger File(s)** (Optional):
Another BOOL format file that is a text-based geometry file used to
add Lead Fingers and Pins that are not present in the design file.
This is only supported in the Python scripting. Format::

    UNITS micron
    NETS YES
    # Comment:
    LEADFRAME LayerName xStart yStart xEnd yEnd
    AUTOPIN pinName

Examples: 'leadfingers.bool'.
Set using the 'leadFiles' attribute.
See `SV_ApexFlow.readLeadFrameFiles`.


**Output Files**:
----------------------

**Project File**:
This is the HyperLynx Advanced Solver (HLAS) project file/folder
that can be opened in HLAS tools (Fast 3D Solver [nApex]).
Example: .phys and associated folder.
Set using the 'projectFile' attribute.


**SPICE File** (Optional):
This is the SPICE file exported from a Fast 3D Solver run (Apex).
Example: 'SPICE.sp'.
Set using the 'exportSPICE' function.


**Port Details File** (Optional):
A file with detailed Port information in XML format::

    <Ports>
        <Port name="<port name>" comp="<component name>" pin="<pin name>" net="<net name>" />
        <Port name="<port name>" comp="<component name>" pin="<pin name>" net="<net name>" />
    </Ports>

Example: 'PortDetails.xml'.
Set using the 'portDetailsFile' attribute.
See `writePortFile`.


**Pin Details File** (Optional):
A file with detailed Pin information in XML format::

    <Pins>
      <Pin name="<name>" comp="<component>" pin="<pinOnly>" net="<net>" layer="<layer>" x="<xCenter(m)>" y="<yCenter(m)>" face="<placement>" />
    </Pins>

Example: 'PinDetails.xml'.
Set using the 'pinDetailsFile' attribute.
See `writePinDetailsFile`.


**Model Details File** (Optional):
A file with detailed Model information in XML format::

    <Model>
        <Nets>
            <Net name="<name>" />
        </Nets>
        <Nodes>
            <Node type="<type>" name="<name>" />
        </Nodes>
    </Model>

Example: 'ModelDetails.xml'.
Set using the 'modelDetailsFile' attribute.
See `writeModelDetailsFile`.


**Renamed Pin Map File** (Optional):
Name of the Pin renaming file. If Pins are renamed,
this file can list the original name and new name.
Example: 'RenamedPinMap.txt'.
Set using the 'pinMapFile' attribute.


.. highlight:: python


Example Design Flow:

>>> import sv_flow
>>> svaf = sv_flow.SV_ApexFlow()
>>> # Design-Specific Options
>>> svaf.inputProjectFile = "./Inputs/Leads.phys"
>>> svaf.projectFile = "Leads.Setup.phys"
>>> svaf.nets = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11"]
>>> svaf.groundNets = ["0"]
>>> svaf.portNets = ".*"
>>> svaf.pinsToMove = [{"pinNames": "EC.*\.[0-9]", "newLayerName": "bottom"}]
>>> # Common Settings (All / Defaults):
>>> svaf.portNaming = "$Net!!$Pin!!$RefComp"
>>> svaf.sinkComponents = "[ULB].*"  # U, L, and B Components are the Sinks
>>> svaf.pinTerminations = True  # Add Pins to Bond Wires and Leads
>>> svaf.groupPins = False  # Group Pins on the same Net & Component (for example, to merge parallel bond wires into the same port)
>>> svaf.pinMapFile = "RenamedPinMap.txt"  # Name of the Pin renaming file
>>> svaf.portDetailsFile = "PortDetails.xml"  # Name of the Port details file
>>> svaf.pinDetailsFile = "PinDetails.xml"  # Name of the Pin details file
>>> svaf.modelDetailsFile = "ModelDetails.xml"  # Name of the Model details file
>>> svaf.spiceSubstitutions = {}  # substitutions to use when exporting the SPICE model
>>> svaf.spiceStages = 1  # (int) Number of stages to use in the exported SPICE file (default 5).
>>> svaf.setLogFile("Messages.log")
>>> svaf.setExtractionType(True, True, 1e8)  # DC, AC, Frequency
>>> svaf.circuits = {'short': 'NT.*'}  # Short all Components starting with NT
>>> # Common Settings (PCB):
>>> svaf.renamePins = False  # flag to control if Pins are renamed
>>> svaf.useCoordinatesInPinName = False  # flag to control if x,y,z are used in Pin Names
>>> svaf.pinsToMove = [{"pinNames": "EC.*\.[0-9]", "newLayerName": "bottom"}]
>>> svaf.solderBalls = ["U1"]
>>> # Common Settings (PKG):
>>> svaf.boolFiles = ["boolfile.bool"]
>>> # Setup, Solve, Export:
>>> svaf.setup()
>>> svaf.projectFile = "Leads.Solved.phys"
>>> svaf.solve()
>>> svaf.exportSPICE("Leads.sp")
>>> svaf.postProcess()
"""

from nimbic import ncore
from nimbictools import flows
from nimbictools import modelutils
from nimbictools import shapeutils
from nimbictools import utils

import math
import os
import re


# Keep a unique version number for this module
version = __version__ = "2017.3.3"


###############################################################################
class SV_ApexFlow(flows.ApexDesignFlow):
    """
    The `SV_ApexFlow` class is used to define the basic
    methods, members, and functionality for a design flow
    catered to SystemVision that uses nApex.

    Inherits all functionality from `flows.ApexDesignFlow` and
    `flows.DesignFlow`.

    All attributes are optional unless noted as **required**.

    **SV-Specific Attributes**:

    Attributes:
      finalNets (str or list of str): list of Net names or a single
        regular expression string to use as the Nets that will be used in the solve.
        Example: "SIG.\*" or ["SIG1", "SIG2"].
      leadFiles (list of str): list of Lead Frame format file names
        to use for adding Leads. see: `readLeadFrameFiles`.
      meshRefinementFactor (int): Mesh Refinement Factor for increasing mesh density.
      modelDetailsFile (str): Name of the Model details file (default "ModelDetails.xml")
      pinDetailsFile (str): Name of the Pin details file (default "PinDetails.xml")
      pinMapFile (str): Name of the Pin renaming file (default "RenamedPinMap.txt")
      portPins (str, list, or dict): An optional dictionary of Pins {source: sink},
        regular expression, or list of Pin names to use for Port Definition.
      portDetailsFile (str): Name of the Port details file (default "PortDetails.xml")
      renamePins (bool): flag to control if Pins are renamed
        ($Component.$Net_x#_y#_z#nm) (default True)
      spiceStages (int): number of stages to use in the exported SPICE file (default 5).
      spiceSubstitutions (dict): substitutions to use when exporting the SPICE model
        (default empty {})
      useBondWireNetNames (bool): Determines if we should use the Bond Wire Net Names from
        the BOOL File or the Net names from the design file (default None: determined by BOOL).
      useCoordinatesInPinName (bool): flag to control if the x,y,z coordinates are used
        in Pin Names (default True)

    **Basic Flow Attributes; sv_flow-Specific Defaults**:

    >>> portNaming = "$Net!!$Pin!!$RefComp"
    >>> portComponents = ".*"  # Use all Components for Ports
    >>> sinkComponents = "[ULB].*"  # U, L, and B Components are the Sinks
    >>> pinTerminations = True  # Add Pins to Bond Wires and Leads
    >>> groupPins = False  # Group Pins on the same Net & Component
    >>> # Set the Solver Options to use Broadband (AC and DC)
    >>> solverOptions = ncore.SolverOptions()
    >>> solverOptions.setExtractionType(True, True, 1e8)
    >>> circuits = {'short': 'NT.*'}  # Short all Components starting with NT

    **Basic Flow Attributes**:

    Attributes:
      tool (str): The tool to use for the flow: nWave, nApex, nSys.
      designFile (str): The Design File to import and simulate.
      inputProjectFile (str): The Project File to read in and setup / process.
        This allows users to save the setup project somewhere other than the original.
        When `inputProjectFile` and `projectFile` are both set,
        the `inputProjectFile` refers to the project to read,
        and `projectFile` is the output file name used to save the project
        after being processed by the flow.
        If only one of them is set, the the project is read in
        and then overwritten by the flow.
      projectFile (str): The Project File to save after setup (and on `save`).
      nets (str or list of str): list of Net names or a single
        regular expression string to use as Nets (Type is inferred).
        Example: "SIG.\*" or ["SIG1", "SIG2"].
      signalNets (str or list of str): list of Net names or a single
        regular expression string to use as Signal Nets.
        Example: "SIG.\*" or ["SIG1", "SIG2"].
      powerNets (str or list of str): list of Net names or a single
        regular expression string to use as Power Nets.
        Example: "VDD.\*" or ["VDD1", "VDD2"].
      groundNets (str or list of str): list of Net names or a single
        regular expression string to use as Ground Nets.
        Example: "VSS.\*" or ["VSS1", "VSS2"].
      portNets (str or list of str): list of Net names or a single
        regular expression string to use for assigning Ports.
        Example: "SIG.\*" or ["SIG1", "SIG2"].
      portComponents (list of str or str or dict): list of Component names
        (e.g. ["BGA1", "U23", ...]), or
        dictionary of Component names with pin pairs
        (e.g. {"BGA1": {"C6": "B4", "B6": "B4", "A6": "GND"},
        "U23": {"24": "2", "25": "2", "26": "GND"}}), or
        a regular expression string (e.g. "BGA.\*")
        of Components to use for assigning Ports.
        Examples: "BGA.\*" or ["BGA1", "U23"] or {"BGA1": {"C6": "B4"}}.
      portReference (str): Reference Net or Layer to use when creating Ports.
      portNaming (str): naming convention to use when creating Ports.
        The following keywords will be replaced by the actual value:
        $Net = Net Name,
        $Comp = Component Name,
        $Pin = Pin Name,
        $RefNet = Reference Net Name,
        $RefComp = Reference Component Name,
        $RefPin = Reference Pin Name.
        If a naming conflict exists, a number will be appended.
      groupPins (bool, str, or list of str): if True, all Pins are automatically grouped by
        Net and Component. If given a list of Net names or a single regular expression string
        all Components are grouped on those Nets specified only.
        Example: True, "VDD.\*", or ["VDD1", "VDD2"].
        For example, to merge parallel bond wires into the same port.
      techFile (str): Technology File for updating the imported Model.
      boolFiles (list of str): list of additional .bool files to read in.
      cropShape (ncore.Shape2D): shape to use when cropping the Model.
      cropNets (str or list of str): list of Net names or a single
        regular expression string to use when cropping by Nets.
        Example: "RX" or ["RX1", "RX2"]
      cropWidth (float): The expansion distance (in meters) to use
        when cropping the Model by Nets.
      groupedNets: a dictionary of NetGroupName to Nets to Group.
        Format: {'<NetGroupName>': <Net Identifier>}
        Example: {'GND': ['Net1', 'Net2']}
        Example: {'GND': 'VSS.\*']}
        Example: {'NetGroupName': ncore.Net.isReference}
        Example: {'NetGroupName': ncore.Net.isGround}
        Use one of the special Net functions to automatically specify nets:
        Identifiers:

        * ncore.Net.isSignal
        * ncore.Net.isPower
        * ncore.Net.isGround
        * ncore.Net.isReference

      solderBalls (list of str or str or list of ncore.SolderBallGroup):
        list of Component names (e.g. ['DIE', 'BGA', ...]), or
        list of SolderBallGroup items (e.g. [ncore.SolderBallGroup(), ...]), or
        a regular expression string (e.g. "BGA.\*")
        of Components to use as Solder Balls.
        The default is to not add Solder Balls.
      leadFrames (list of str or str or list of ncore.LeadFrameGroup):
        list of layer names (e.g. ['PKG', 'LF', ...]), or
        list of LeadFrameGroup items (e.g. [ncore.LeadFrameGroup(), ...]), or
        a regular expression string for Layer names (e.g. "PKG.\*")
        for adding Lead Frames.
        The default is to not add Lead Frames.
      floodedReferences (list of str or str or dict):
        list of Bonding Group names (e.g. ['DIE', 'BGA', ...]), or
        a regular expression string (e.g. "BGA.\*")
        of Bonding Groups to use for flooded references, or
        a Dictionary of the Name of the Net to use as the
        Reference Net to the Bonding Group Names:
        e.g. {"GND": ["DIE1", "DIE2"]: "VSS": "BGA.\*"}.
        The default is to not add Flooded References.
      componentsToMove (dict): Dictionary of Component Names to list of:
        Layer Name and Placement for moving Pins.
        Format: {"CompName": ["LayerName", "PinPlacement"], ...}
      componentsToDelete (list of str or str): list of Component names
        (e.g. ["BGA1", "U23", ...]), or
        a regular expression string (e.g. "BGA.\*")
        of Components to delete after importing.
        Examples: "BGA.\*" or ["BGA1", "U23"].
      pinsToMove (list of dict): List of Pin Names, Layer Name, and Placement for moving Pins.
        Format: [{"pinNames": "Pins", "newLayerName": "LayerName",
        "newPlacement": "PinPlacement", ...}, ...]
        See `modelutils.changePinPlacement` for the full list.
      pinTerminations (bool): if True, Pins are added to the Termination side of Bond Wires,
        Solder Balls, and Lead Frames (default False).
      circuits (dict): Dictionary of Circuit model Names to Component Names.
        Format: {"<circuitName>": ["<ComponentName1>", "<ComponentName2>"]}
        Format: {"<circuitName>": "<ComponentRegularExpression>"}
        e.g. {"open": ["C1", "C2"]}
        e.g. {"short": "NT.\*"}
      removeBrokenNets (bool): if True, Nets that are broken will be split and
        pieces without pins will be removed. Default False.
      shapeVariability (list of `shapeutils.ShapeVariabilityOptions`):
        see `modelutils.expandShapesByFactor` and `designutils.createShapeVariability` and
        `shapeutils.ShapeVariabilityOptions` for more details.
      meshOptions (ncore.MeshOptions): Mesh Options for the Simulation.
      solverOptions (ncore.SolverOptions): Solver Options for the Simulation.
      layoutOptions (ncore.LayoutOptions): Layout Options for the Simulation.
      arcSegments (int): The minimum number of segments used in 360 degrees of arc.
      arcDeviation (float): The maximum distance allowed between
        the 'true arc' and the segmented arc.
      report (reportutils.NimbicReport): Report to generate for the Flow.
      saveImages (bool): Determines whether or not to save Model Image files during a save.

    """

    def __init__(self):
        """
        Initialization for the SV_ApexFlow class.
        """
        super(SV_ApexFlow, self).__init__()
        self.portNaming = "$Net!!$Pin!!$RefComp"
        # self.portComponents = ".*"
        self.sinkComponents = "[UL].*"  # U and L Components are the Sinks
        self.pinsToMove = None
        self.pinTerminations = False
        self.groupPins = None
        # Set the Solver Options to use Broadband (AC and DC)
        self.solverOptions = ncore.SolverOptions()
        self.solverOptions.setExtractionType(True, True, 1e8)
        self.circuits = {'short': 'NT.*'}  # Short all Components starting with NT
        self._unlock_()
        self.finalNets = None
        self.leadFiles = None
        self.pinDetailsFile = "PinDetails.xml"
        self.pinMapFile = "RenamedPinMap.txt"
        self.portDetailsFile = "PortDetails.xml"
        self.modelDetailsFile = "ModelDetails.xml"
        self.portPins = None
        self.renamePins = True
        self.spiceStages = 5
        # self.spiceSubstitutions = {'+': 'p', '-': 'n', '<EMPTY_NET>': 'EMPTY_NET'}
        self.spiceSubstitutions = None
        self.useBondWireNetNames = None
        self.useCoordinatesInPinName = True
        self.leadSnapPoints = None
        self.meshRefinementFactor = None
        self._lock_()
        pass

    # ###########################################################
    # Overridden ApexDesignFlow Methods:
    def postImportOperations(self):
        """
        A place to run various model operations after Importing.
        """
        super(SV_ApexFlow, self).postImportOperations()
        # Determine if we should use the Bond Wire Net Names from the BOOL File:
        if self.useBondWireNetNames is None and self.boolFiles:
            for bf in self.boolFiles:
                with open(bf) as f:
                    lines = f.readlines()
                    f.close()
                for line in lines:
                    parts = line.split()
                    if not parts:
                        continue
                    command = parts[0].upper()
                    if command == "NETS" and len(parts) > 1:
                        self.useBondWireNetNames = parts[1].upper().startswith("Y")
        if self.useBondWireNetNames:
            updateNetsFromBondWires(self.model)
        else:
            updateBondWireNets(self.model)
        # self.updateShapeNets()
        self.save()
        pass

    def preCropOperations(self):
        """
        A place to run various model operations before cropping.
        """
        super(SV_ApexFlow, self).preCropOperations()
        self.leadSnapPoints = readLeadFrameFiles(self.model, self.leadFiles)
        pass

    def postCropOperations(self):
        """
        A place to run various model operations after cropping.
        """
        super(SV_ApexFlow, self).postCropOperations()
        pass

    def prePortAssignmentOperations(self):
        """
        A place to run various model operations before assigning ports.
        """
        super(SV_ApexFlow, self).prePortAssignmentOperations()
        self.model.createValidNets(True)
        self.assignPorts()
        checkForMissingBondWirePins(self.model)
        modelutils.deleteComponents(self.model, componentNames=['PKG_TO_PKG'])
        # Rename the Pins:
        if self.renamePins:
            renameModelPins(self.model,
                            mapFile=self.pinMapFile,
                            useCoordinates=self.useCoordinatesInPinName)
        self.save()
        pass

    def assignCircuits(self):
        """
        A place to assign circuit ports before assigning ports.
        """
        if self.circuits and utils.isType(self.circuits, dict):
            # Delete 0-Pin Components and 1-Pin Components with Circuits:
            modelutils.deleteComponents(self.model, pinCount=lambda x: x < 1)
            for key, value in self.circuits.items():
                modelutils.deleteComponents(
                    self.model, pinCount=lambda x: x < 2, componentNames=value)
        self.save()
        super(SV_ApexFlow, self).assignCircuits()
        pass

    def assignPorts(self):
        """
        Determines the method for assigning Ports.

        """
        if self.groupPins:  # Group the Pins
            if utils.isType(self.groupPins, bool):
                self.model.autoGroupPins()
            else:
                modelutils.groupPinsOnSameComponent(self.model, self.groupPins)
            self.save()
        if utils.isType(self.portPins, dict):
            self.assignPortsByPin()
        else:
            self.assignPortsByNet()
        pass

    def assignPortsByNet(self):
        """
        Assign Ports: Port Definition loops over every Net with 2+ Pins,
        finds a valid Sink Pin, and creates a Port.

        Uses these attributes of the class:

        * `groupPins`
        * `portNets`
        * `portComponents`
        * `portNaming`

        """
        modelutils.assignPortsByNet(
            self.model, portNets=self.portNets,
            portComponents=self.portComponents, portPins=self.portPins,
            sinkComponents=self.sinkComponents, portNaming=self.portNaming,
            groupPins=False)  # groupPins=self.groupPins)
        self.save()
        pass

    def assignPortsByPin(self):
        """
        Assign Ports: Port Definition loops over every Net with 2+ Pins,
        finds a valid Sink Pin, and creates a Port.

        Uses these attributes of the class:

        * `portPins`
        * `portNaming`

        Example Usage:

        >>> import sv_flow
        >>> svaf = sv_flow.SV_ApexFlow()
        >>> svaf.designFile = "ODB_TC2_eval_rev0.zip"
        >>> svaf.projectFile = "portPins.phys"
        >>> svaf.nets = ["VIN", "VOUT"]
        >>> svaf.renamePins = False
        >>> svaf.portPins = {"EC1.K": "U1.4", "EC1.B": "VOUT1.1"}
        >>> svaf.setup()

        """
        self.save()
        po = self.model.getPortOptions()
        md = modelutils.getModelDetails(self.model)
        portNaming = self.portNaming
        if not portNaming:
            portNaming = "$Net!!$Pin!!$RefComp"
        if not utils.isType(self.portPins, dict):
            raise RuntimeError("Invalid 'portPins' value. Format is {'sourcePin': 'sinkPin'}")
        for sourcePinName, sinkPinName in self.portPins.items():
            sourcePin = md.getPin(sourcePinName)
            sinkPin = md.getPin(sinkPinName)
            sourceNetName = sourcePin.getNetName()
            sinkNetName = sinkPin.getNetName()
            name = portNaming.replace("$Net", sourceNetName)\
                .replace("$Pin", sourcePinName)\
                .replace("$RefPin", sinkPinName)\
                .replace("$RefNet", sinkNetName)\
                .replace("$Comp", sourcePin.getComponentName())\
                .replace("$RefComp", sinkPin.getComponentName())
            port = ncore.PortDescription()
            port.setName(name)
            port.setSourceNetName(sourceNetName)
            port.setSinkNetName(sinkNetName)
            port.setSourcePinName(sourcePinName)
            port.setSinkPinName(sinkPinName)
            self.model.displayInformation("Adding Port: %s" % port)
            po.addPort(port)
        # Sort all Ports by Name
        po.sortChildrenByName()
        self.model.assignPorts(po)
        self.save()
        pass

    def postPortAssignmentOperations(self):
        """
        A place to run various model operations after assigning ports.
        """
        super(SV_ApexFlow, self).postPortAssignmentOperations()
        self.save()
        if self.portDetailsFile:
            writePortFile(self.portDetailsFile, self.model)
        if self.pinDetailsFile:
            writePinDetailsFile(self.pinDetailsFile, self.model, leadSnapPoints=self.leadSnapPoints)
        if self.modelDetailsFile:
            writeModelDetailsFile(self.modelDetailsFile, self.model)
        if self.finalNets:
            modelutils.keepNets(self.model, self.finalNets)
        pass

    def checkValidity(self):
        """
        Checks for some common errors in the Model setup.
        """
        md = modelutils.getModelDetails(self.model)
        errors = ""
        nets = md.getNets()
        for net in nets:
            # Check Nets for 1 Pin Only:
            pins = net.getPins()
            if len(pins) == 1:
                errors += "Net '{}' is invalid, since it has only 1 Pin '{}'.\n"\
                    .format(net.getName(), pins[0].getName())
        # if errors:
        #     raise RuntimeError(errors)

        MeshRefinement.autoAdjustRefinement(self.simulation, self.meshRefinementFactor)
        super(SV_ApexFlow, self).checkValidity()
        pass

    def getMemoryEstimate(self):
        """
        Returns the memory estimate as a string with MB or GB (example: "255MB")

        Example:
          Projected Memory Requirement: ~255 MB.

        Returns:
          str: memory estimate (example: "255MB")
        """
        simulations = self.project.getSimulations()
        memory = ""
        for sim in simulations:
            if not sim.doesMeshExist():
                sim.mesh()
            # memoryString is a Python string and can be parsed for the memory
            memoryString = sim.estimateMemory()
            print(memoryString)
            if "~" in memoryString:
                memory = str(memoryString)
                s = memory.index("~")
                memory = memory[s + 1:].replace("B.", "B").replace(" ", "")
        return memory

    def exportSPICE(self, spiceFileName, modelType="PI",
                    netlistNamePrefix=None, replacementSubckt=None, portMapFile=None,
                    optionsFile=None):
        """
        Export the simulation SPICE file.

        Args:
          spiceFileName (str): the SPICE File Name to export to.
          modelType (str): "PI" or "L" [Default "PI"].
          netlistNamePrefix (str): the name to use for the netlist Prefix.
          replacementSubckt (str): the name to use for the top circuit.
          portMapFile (str): the input Map file for renaming the Nodes.
          optionsFile (str): the Option File to use when writing the SPICE Model.

        Returns:
          str: full path name to the exported SPICE File.
        """
        return exportSPICE(self.simulation, spiceFileName,
                           spiceStages=self.spiceStages,
                           modelType=modelType,
                           spiceSubstitutions=self.spiceSubstitutions,
                           netlistNamePrefix=netlistNamePrefix,
                           replacementSubckt=replacementSubckt,
                           portMapFile=portMapFile,
                           optionsFile=optionsFile)

    pass


class SV_WaveFlow(flows.WaveDesignFlow):
    """
    The `SV_WaveFlow` class is used to define the basic
    methods, members, and functionality for a design flow
    catered to SystemVision that uses nWave.

    Inherits all functionality from `flows.WaveDesignFlow` and
    `flows.DesignFlow`.

    All attributes are optional unless noted as **required**.

    Attributes:

    """

    def __init__(self):
        """
        Initialization for the SV_WaveFlow class.
        """
        super(SV_WaveFlow, self).__init__()
        self._unlock_()
        self._lock_()
        pass

    def prePortAssignmentOperations(self):
        super(SV_WaveFlow, self).prePortAssignmentOperations()
        self.setFlatteningMacroModelOptions()
        pass

    def setFlatteningMacroModelOptions(self):
        simulations = self.project.getSimulations()
        flattening = 'NO FLATTEN'
        macromodel = 'NO MACROMODEL'
        for sim in simulations:
            model = sim.getModel()
            modelutils.setFlatteningMacroModelOptions(model, flattening, macromodel)
        pass

    pass


class SV_HybridFlow(flows.SysSIDesignFlow):
    """
    The `SV_HybridFlow` class is used to define the basic
    methods, members, and functionality for a design flow
    catered to SystemVision that uses nSys for Signal Integrity.

    Inherits all functionality from `flows.SysSIDesignFlow` and
    `flows.DesignFlow`.

    All attributes are optional unless noted as **required**.

    Attributes:

    """

    def __init__(self):
        """
        Initialization for the SV_WaveFlow class.
        """
        super(SV_HybridFlow, self).__init__()
        self._unlock_()
        self._lock_()
        pass

    pass


################################################################################
# Custom Model Methods:
def updateBondWireNets(model):
    """
    Update the Bond Wire Net Information to match the geometry.
    """
    update = False
    md = modelutils.getModelDetails(model)
    bwgs = md.getBondWireGroups()
    for bwg in bwgs:
        bwgn = bwg.getName()
        supportLayerName = bwg.getSupportLayerName()
        bws = bwg.getBondWires()
        n = 1
        for bw in bws:
            # bw.setNetName("<EMPTY_NET>")
            net = model.queryNetOnLayer(bw.getStartPoint(), supportLayerName)
            if not net:
                net = bw.getNetName()
            if not net:
                net = "%s.BW%d" % (bwgn, n)
                n += 1
            bw.setNetName(net)
            print(str(bw))
            update = True
    if update:
        model.setDetails(md)
        model.createValidNets(True)
    pass


def updateNetsFromBondWires(model):
    """
    Update the Net Information from the given Bond Wire Net Information.
    """
    update = False
    md = modelutils.getModelDetails(model)
    bwgs = md.getBondWireGroups()
    newToOld = {}
    oldToNew = {}
    for bwg in bwgs:
        supportLayerName = bwg.getSupportLayerName()
        bws = bwg.getBondWires()
        for bw in bws:
            oldNet = model.queryNetOnLayer(bw.getStartPoint(), supportLayerName)
            newNet = bw.getNetName()
            if not oldNet or not newNet:
                continue
            if oldNet in oldToNew:
                tempNet = oldToNew[oldNet]
                oldNet = newNet
                newNet = tempNet
            if newNet in newToOld:
                newToOld[newNet].append(oldNet)
            else:
                newToOld[newNet] = [oldNet]
            update = True
            oldToNew[oldNet] = newNet
    if update:
        modelutils.groupNets(model, newToOld)
        model.createValidNets(True)
    pass


def updateShapeNets(self):
    """
    Update the Shape Net Information to match the geometry.
    """
    update = False
    md = modelutils.getModelDetails(self.model)
    vias = md.getVias()
    for via in vias:
        via.setNetName("<EMPTY_NET>")
        update = True
    shapes = md.getShapes()
    for shape in shapes:
        shape.setNetName("<EMPTY_NET>")
        update = True
    if update:
        self.model.setDetails(md)
        self.model.createValidNets(True)
    pass


def checkForMissingBondWirePins(model):
    """
    Check for 'floating' Bond Wires report errors.
    """
    # Find out which Ports were added:
    po = model.getPortOptions()
    netsWithPorts = set()
    netsWithoutPorts = set()
    bondWireGroupsWithoutPorts = set()
    for port in po.getPorts():
        netsWithPorts.add(port.getSourceNetName())
    md = modelutils.getModelDetails(model)
    # Check for 'floating' Bond Wires:
    errors = ""
    bwgs = md.getBondWireGroups()
    for bwg in bwgs:
        bwgName = bwg.getName()
        bws = bwg.getBondWires()
        i = 0
        for bw in bws:
            i += 1
            net = bw.getNetName()
            if net in netsWithPorts:
                continue
            # 'Floating' Bond Wire found:
            netsWithoutPorts.add(net)
            bondWireGroupsWithoutPorts.add(bwgName)
            errors += "Bond Wire #{} in Group '{}' on Net '{}' missing Pin.\n".\
                format(i, bwgName, net)
            errors += "  Ensure its Pin is on the Bond Wire " \
                      "Support / Termination Layer and the same Net.\n"
    if errors:
        raise RuntimeError(errors)
    # Add New Pins to 'floating' Bond Wires, if needed:
    if bondWireGroupsWithoutPorts and netsWithoutPorts:
        model.createSpecificPinTerminations(list(bondWireGroupsWithoutPorts), True, False)
        # msg = "The Bond Wires in the Groups '{}'".format(str(bondWireGroupsWithoutPorts))
        # msg += "are missing Ports on the Nets '{}'".format(str(netsWithoutPorts))
        # raise RuntimeError(msg)
    # Clear the Ports:
    po = ncore.PortOptions()
    model.assignPorts(po)
    pass


def readLeadFrameFiles(model, leadFiles):
    """
    Read the Lead Finger placements from files.

    Format::

        UNITS unitstring
        LEADFRAME LayerName xStart yStart xEnd yEnd
        AUTOPIN PinName

    Example::

        UNITS micron
        LEADFRAME LF1 -2385.0 -2258.8 -2165.0 -2258.8
        AUTOPIN PIN_LF1


    Args:
      model (ncore.Model): The Geometry Model.
      leadFiles (list of str): list of Lead Frame format file names to use for adding Leads.
    """
    leadSnapPoints = []
    if not leadFiles:
        return leadSnapPoints
    # if not utils.isType(self.leadFiles, list):
    #     self.leadFiles = [self.leadFiles]
    # Fix a problem with Finalize
    md = model.getDetails()
    model.setDetails(md)
    if shapeutils.isLayerShapeVersion():
        model.applyPads()
    else:
        model.finalize()  # Needed to convert Vias to Shapes for the valid net creation
    md = modelutils.getModelDetails(model)
    unit = md.getLengthUnit()
    md.setLengthUnit("m")
    lfgs = md.getLeadFrameGroups()
    if len(lfgs) < 1:
        return leadSnapPoints
    for lfg in lfgs:
        lfg.deleteAllChildren()
        lfg.setModifiedFlag(True)
    line, lineFormat, leadFile = "", "", ""
    lineNumber = 0
    modified = False
    try:
        unitMult = 1.0
        for leadFile in leadFiles:
            if not utils.isTypes(leadFile, [str, unicode]):
                raise RuntimeError("Invalid Lead Frame File Name '{}'".format(leadFile))
            utils.checkFileExists(leadFile)
            with open(leadFile) as f:
                lines = f.readlines()
            if not lines:
                continue
            nets = False
            lineNumber = 0
            pin, lf, leadFrameGroup = None, None, None
            while lineNumber < len(lines):
                line = lines[lineNumber]
                lineNumber += 1
                parts = line.split()
                if not parts:
                    continue
                command = parts[0].upper()
                if command == "LEADFRAME":
                    lineFormat = "LEADFRAME LayerName xStart yStart xEnd yEnd"
                    lfgName = parts[1]
                    leadFrameGroup = None
                    for lfg in lfgs:
                        if lfg.getName() == lfgName:
                            leadFrameGroup = lfg
                            break
                    if not leadFrameGroup:
                        raise RuntimeError("Invalid Lead Frame Group Name '{}'".format(lfgName))
                    lf = ncore.LeadFrame()
                    lf.setLayerName(lfgName)
                    coords = parts[2:]
                    net = "<EMPTY_NET>"
                    if nets:
                        lineFormat = "LEADFRAME LayerName NetName xStart yStart xEnd yEnd"
                        net = parts[2]
                        coords = parts[3:]
                    coords = [float(c) for c in coords]
                    xs, ys, xe, ye = coords[0], coords[1], coords[2], coords[3]
                    # Snap the Start and End Points to the Closest Points on the Layer:
                    # Get the Snap Points from the Support Layer:
                    layers = modelutils.getFilteredLayers(
                        md, layerNames=[leadFrameGroup.getSupportLayerName()])
                    snapPoints = getLayersPoints(layers)
                    sp = ncore.Point2D(xs * unitMult, ys * unitMult)
                    ep = ncore.Point2D(xe * unitMult, ye * unitMult)
                    lf.setStartPoint(sp)
                    lf.setEndPoint(ep)
                    pt = ncore.Point2D.getMiddlePoint(sp, ep)
                    leadSnapPoints.append(pt)
                    sp = getClosestPoint(snapPoints, sp)
                    ep = getClosestPoint(snapPoints, ep)
                    # If the Snap makes the Start and End Points too close, don't apply it
                    if sp.getDistanceTo(ep) > 1e-10:
                        lf.setStartPoint(sp)
                        lf.setEndPoint(ep)
                    if not nets:
                        # Query the Net where the Lead Finger touches
                        supportLayer = leadFrameGroup.getSupportLayerName()
                        xs, ys = lf.getStartPoint().x(), lf.getStartPoint().y()
                        xe, ye = lf.getEndPoint().x(), lf.getEndPoint().y()
                        xc, yc = 0.5 * (xs + xe), 0.5 * (ys + ye)
                        # Shift it slightly:
                        shift = 0.1 * math.sqrt((xe - xs) ** 2 + (ye - ys) ** 2)
                        a = math.pi / 2.0 - math.atan2(ye - ys, xe - xs)
                        xc -= shift * math.cos(a)
                        yc += shift * math.sin(a)
                        pt = ncore.Point2D(xc, yc)
                        net = model.queryNetOnLayer(pt, supportLayer)
                        if not net:
                            # Shift it back:
                            xc += shift * math.cos(a)
                            yc -= shift * math.sin(a)
                            pt = ncore.Point2D(xc, yc)
                            net = model.queryNetOnLayer(pt, supportLayer)
                        if not net:
                            # Shift it more:
                            xc += shift * math.cos(a)
                            yc -= shift * math.sin(a)
                            pt = ncore.Point2D(xc, yc)
                            net = model.queryNetOnLayer(pt, supportLayer)
                    lf.setNetName(net)
                    leadFrameGroup.addChild(lf)
                    modified = True
                elif command == "PIN":
                    lineFormat = "PIN PinName\nCOMPONENT_NAME CompName\n" \
                                 "RECTANGLE LayerName ADD xCenter yCenter width height angle"
                    if nets:
                        lineFormat = \
                            "PIN PinName\nCOMPONENT_NAME CompName\n" \
                            "RECTANGLE LayerName NetName ADD xCenter yCenter width height angle"
                    if len(parts) < 2:
                        raise RuntimeError("Invalid Pin Defined")
                    pin = ncore.Pin()
                    pin.setPinName(parts[1])
                elif command == "COMPONENT_NAME":
                    if len(parts) < 2 or not pin:
                        raise RuntimeError("Invalid Pin Defined")
                    pin.setComponentName(parts[1])
                elif command == "RECTANGLE":
                    if len(parts) < 8 or not pin:
                        raise RuntimeError("Invalid Pin Defined")
                    pin.setLayerName(parts[1])
                    coords = parts[3:]
                    if nets:
                        pin.setNetName(parts[2])
                        coords = parts[4:]
                    else:
                        pin.setNetName("<EMPTY_NET>")
                    coords = [float(c) for c in coords]
                    xc, yc, w, h = coords[0], coords[1], coords[2], coords[3]
                    a = 0
                    if len(coords) > 4:
                        a = coords[4]
                    rect = ncore.Rectangle2D()
                    rect.setCenter(ncore.Point2D(xc * unitMult, yc * unitMult))
                    rect.setWidth(w * unitMult)
                    rect.setHeight(h * unitMult)
                    rect.setAngle(a)
                    pin.setShape(rect)
                    pin.setPlacement("BOTTOM")
                    pin.setModifiedFlag(True)
                    md.addPin(pin)
                    modified = True
                elif command == "AUTOPIN":
                    lineFormat = "AUTOPIN PinName"
                    if len(parts) < 2:
                        raise RuntimeError("Invalid Auto Pin Defined.")
                    pin = ncore.Pin()
                    pinName = parts[1]
                    pin.setPinName(pinName)
                    if not lf or not leadFrameGroup:
                        raise RuntimeError("Invalid Lead Frame / Pin '{}' Defined.".format(pinName))
                    pin.setComponentName(leadFrameGroup.getName())
                    pin.setLayerName(leadFrameGroup.getName())
                    rect = ncore.Rectangle2D()
                    xs, ys = lf.getStartPoint().x(), lf.getStartPoint().y()
                    xe, ye = lf.getEndPoint().x(), lf.getEndPoint().y()
                    xc, yc = 0.5 * (xs + xe), 0.5 * (ys + ye)
                    h = math.sqrt((xe - xs) ** 2 + (ye - ys) ** 2)
                    w = leadFrameGroup.getLength() * leadFrameGroup.getBottomFactor()
                    cc = leadFrameGroup.getLength() - 0.5 * w
                    a = math.atan2(ye - ys, xe - xs)
                    a = math.pi / 2.0 - a
                    xc += cc * math.cos(a)
                    yc -= cc * math.sin(a)
                    rect.setCenter(ncore.Point2D(xc, yc))
                    if h < 10e-6 or w < 10e-6:  # Give an error for tiny Pins
                        raise RuntimeError(
                            "Invalid Pin '{}' due to invalid height or width "
                            "for Lead Frame Group '{}'.".format(pinName, leadFrameGroup.getName()))
                        # continue
                    rect.setWidth(w)
                    rect.setHeight(h)
                    rect.setAngle(a * -180.0 / math.pi)
                    pin.setShape(rect)
                    net = lf.getNetName()
                    if not net:
                        raise RuntimeError(
                            "Invalid Pin '{}' due to empty Lead Net.".format(pinName))
                        # continue
                    pin.setNetName(lf.getNetName())
                    pin.setPlacement("BOTTOM")
                    pin.setModifiedFlag(True)
                    md.addPin(pin)
                    modified = True
                elif command == "NETS":
                    lineFormat = "NETS YES"
                    nets = parts[1].upper().startswith("Y")
                elif command == "UNITS":
                    lineFormat = "UNITS micron"
                    unitMult = utils.getLengthUnitMultiplier(parts[1])
    except Exception as e:
        raise RuntimeError(
            "Error reading line '{}'\nin file '{}' near line# {}.\nFormat: {}\n{}"
            .format(line.strip(), leadFile, lineNumber, lineFormat, str(e)))
    md.setLengthUnit(unit)
    if modified:
        model.displayInformation(str(md))
        for pin in md.getPins():
            model.displayInformation("Adding Pin to Net '{}' on Layer '{}'".format(
                pin.getNetName(), pin.getLayerName()))
        for lfg in md.getLeadFrameGroups():
            for lf in lfg.getLeadFrames():
                model.displayInformation("Adding Lead to Net '{}' on Layer '{}'".format(
                    lf.getNetName(), lf.getLayerName()))
        model.setDetails(md)
        model.createValidNets(True)
    return leadSnapPoints


def snapLeadFingerPoints(model):
    """
    Snap the Lead Finger points to the edges of metal.

    Args:
      model (ncore.Model): The Geometry Model.
    """
    md = modelutils.getModelDetails(model)
    unit = md.getLengthUnit()
    md.setLengthUnit("m")
    lfgs = md.getLeadFrameGroups()
    if len(lfgs) < 1:
        return
    modified = False
    for lfg in lfgs:
        lfg.setModifiedFlag(True)
        lfs = lfg.getLeadFrames()
        for lf in lfs:
            # Snap the Start and End Points to the Closest Points on the Layer:
            # Get the Snap Points from the Support Layer:
            layers = modelutils.getFilteredLayers(
                md, layerNames=[lfg.getSupportLayerName()])
            snapPoints = getLayersPoints(layers)
            sp = lf.getStartPoint()
            ep = lf.getEndPoint()
            sp = getClosestPoint(snapPoints, sp)
            ep = getClosestPoint(snapPoints, ep)
            # If the Snap makes the Start and End Points too close, don't apply it
            if sp.getDistanceTo(ep) > 1e-10:
                lf.setStartPoint(sp)
                lf.setEndPoint(ep)
            modified = True
    md.setLengthUnit(unit)
    if modified:
        model.setDetails(md)
    pass


def convertSpiceToVirtuoso(inSpiceFile, outSpiceFile, substitution=None,
                           mapFile=None, nodePrefix=None, elementPrefixes=None,
                           netlistNamePrefix=None, replacementSubckt=None, portMapFile=None):
    """
    ;;------------------------------------------------------------------------------
    ;; TexasInstruments, Inc. Copyright 2015                                     --
    ;;------------------------------------------------------------------------------
    ;; Description    : Python utility for post processing a spice file generated --
    ;; by Nimbic nApex for compatibility with Cadence Virtuoso                    --
    ;; Creation Date  : 2Q 2015                                                   --
    ;; Author(s)      : Steven Tanner                                             --
    ;; Contributor(s) : Devan Williams (devan_williams@mentor.com)                --
    ;;------------------------------------------------------------------------------
    ;; Copyright (c) 2015 TexasInstruments Inc.                                  --
    ;; This  is  an  unpublished   work   created  in  the  year   stated   above --
    ;; Texas  Instruments owns  all  rights in  and to this  work and  intends to --
    ;; maintain  and  protect  it as an  unpublished  copyright. In the  event of --
    ;; either inadvertent or deliberate publication,  the above stated date shall --
    ;; treated as the year of first publication. In the event of such publication --
    ;; Texas  Instruments  intends  to  enforce its rights in  the work under the --
    ;; copyright laws as a published work.                                        --
    ;;------------------------------------------------------------------------------

    Args:
      inSpiceFile (str): input SPICE file name (**required**).
      outSpiceFile (str): output SPICE file name (**required**).
      substitution (dict): the substitution for the mapping control.
        This example would substitute any '+' characters to 'p' and any '-' characters to 'n':
        (e.g. "{'+': 'p', '-': 'n'}")
      mapFile (str): file name of the map file with original names to mapped names.
      nodePrefix (str): character to use when prefixing numbers (default is 'a').
      elementPrefixes (list of str): the prefixes of the Elements for node-renaming.
        (e.g. ['R', 'L', 'C'] default is ['V', 'K', 'R', 'L', 'G', 'C', 'I']).
      netlistNamePrefix (str): the name to use for the netlist Prefix.
      replacementSubckt (str): the name to use for the top circuit.
      portMapFile (str): the input Map file for renaming the Nodes.

    Example:

    >>> convertSpiceToVirtuoso('Before.sp', 'After.sp', {'+': 'p', '-': 'n'}, 'netnames.map')

    """
    # Build port dictionary
    outtext = ""
    maptext = ""
    ports = {}
    subckts = []
    fflag = -1
    renameMap = {}
    if not nodePrefix:
        nodePrefix = "a"
    if not elementPrefixes:
        elementPrefixes = ["V", "K", "R", "L", "G", "C", "I"]
    try:
        if portMapFile and os.path.isfile(portMapFile) and os.access(portMapFile, os.R_OK):
            with open(portMapFile) as PORTS:
                for line in PORTS:
                    line = line.replace("\n", "")
                    (key, val) = line.split(' ')
                    ports[key.strip()] = val.strip()

        utils.checkFileExists(inSpiceFile)
        with open(inSpiceFile) as IN:
            inLines = IN.readlines()
        maptext += "Original      Mapped Name\n"
        maptext += "----------    -------------\n"
        origSubckt = ""
        for line in inLines:
            words = line.split()
            if len(words) > 5 and words[1] == "SPICE":
                if words[2] == "Node":
                    origWord = words[5]
                    newWord = words[5]
                    if substitution:
                        for key, value in substitution.items():
                            newWord = newWord.replace(r'%s' % key, value)
                    if origWord in renameMap.keys():
                        raise RuntimeError("Error!!! Duplicate port '%s' found in in file '%s': "
                                           % (origWord, inSpiceFile))
                    else:
                        renameMap[origWord] = newWord
                        maptext += r'%s = %s' % (origWord, newWord)
                        maptext += '\n'
            # Find all Subcircuit Names and
            # Force .ends statement for main subckt:
            if len(words) > 1 and words[0] == ".subckt":
                fflag += 1
                subckt = words[1]
                if fflag == 0:
                    origSubckt = subckt
                    if replacementSubckt:
                        line = line.replace(subckt, replacementSubckt)
                        subckt = replacementSubckt
                elif fflag == 1:
                    if replacementSubckt:
                        outtext += r".ends %s" % replacementSubckt
                    else:
                        outtext += r".ends %s" % subckt
                    outtext += "\n\n"
                subckts.append(subckt)
            # Substitute ports and nets which contain illegal characters
            for port in ports:
                line = re.sub(r"\b%s\b" % port, ports[port], line)
            for key, value in renameMap.items():
                line = line.replace(r'%s' % key, value)
            # Add internal net prefix if missing
            words = line.split()
            if line[:1] in elementPrefixes:
                outtext += words[0] + " "
                outtext += " ".join([nodePrefix + w if w.isdigit() else w for w in words[1:3]])
                outtext += " " + " ".join(words[3:]) + "\n"
            else:
                # Do not print .ends of the main subckt at end of file
                if len(words) > 1 and fflag > 0 and words[1] == origSubckt:
                    continue
                outtext += " ".join([nodePrefix + w if w.isdigit() else w for w in words])
                outtext += "\n"
        if mapFile:
            with open(mapFile, 'w') as MAP:
                MAP.write(maptext)
                MAP.close()
        with open(outSpiceFile, 'w') as OUT:
            OUT.write(outtext)
            OUT.close()
        if netlistNamePrefix:
            with open(outSpiceFile) as IN:
                inLines = IN.readlines()
                for i in range(len(inLines)):
                    words = inLines[i].split()
                    if len(words) > 1:
                        w0 = words[0]
                        w1 = words[1]
                        if (w0 == ".subckt" or w0 == ".ends") or (w0 == "+" and w1 in subckts):
                            inLines[i] = "%s %s%s\n" % (w0, netlistNamePrefix, w1)
                IN.close()
            with open(outSpiceFile, 'w') as OUT:
                OUT.write("".join(inLines))
                OUT.close()
    finally:
        pass
    pass


def deleteUnwantedLeadsOutsideShape(model, shape):
    """
    Delete Unwanted Leads in a Model (by Shape).
    """
    modelutils.deleteLeadsOutsideShape(model, shape)
    pass


def deleteUnwantedLeadsOutsidePoints(model, points):
    """
    Delete Unwanted Leads in a Model (by Polygon Points).
    Input should be in meters.
    """
    modelutils.deleteLeadsOutsidePoints(model, points)
    pass



def ensureUniqueName(name, names):
    """
    Ensure the given name is unique.
    """
    n = 1
    newName = name
    baseName = str(newName)
    while newName in names:
        newName = baseName + "_" + str(n)
        n += 1
    names.append(newName)
    return newName


def getNewPinName(pin, useCoordinates, names):
    """
    Return the new Pin Name from the Pin and format, ensuring unique.

    Args:
      pin (ncore.Pin): Pin for which to get the name.
      useCoordinates (bool): append the x,y,z coordinates or not.
      names (list of str): names to ensure uniqueness.
    """
    newName = "%s.%s" % (pin.getComponentName(), pin.getNetName())
    if useCoordinates:
        bb = pin.getBoundingBox()
        x, y, z = bb.getCenterXYZ()
        newName += "_x%d_y%d_z%dnm" % (x * 1e9, y * 1e9, z * 1e9)
    newName = ensureUniqueName(newName, names)
    return newName


def renameModelPins(model, mapFile=None, useCoordinates=True, adjustLeads=True):
    """
    Renames the Pins as 'Component.Net_x####_y####_z####'.

    Ensures the names are unique by appending '_#' until the name does not exist.

    Args:
      model (ncore.Model): The Geometry Model.
      mapFile (str): file name for saving the old pin name to new pin name map.
      useCoordinates (bool): append the x,y,z coordinates or not.
      adjustLeads (bool): check if we should adjust the lead pin names.
    """
    md = modelutils.getModelDetails(model)
    ro = ncore.RenameOptions()
    ro.setType("Pins")
    text = "Old Name = New Name\n"
    renamedPins = []
    names = []
    pins = md.getPins()
    if useCoordinates and adjustLeads:
        # Lead Pins Renamed based on Start Point
        md.setLengthUnit("m")
        lfgs = md.getLeadFrameGroups()
        for lfg in lfgs:
            lfs = lfg.getLeadFrames()
            for lf in lfs:
                # print lf
                utils.checkType(lf, ncore.LeadFrame)
                lfbb = lf.getBoundingBox()
                pin = lf.getTerminationPin()
                if not pin:
                    # Look for the Pin in the List of Pins:
                    x, y, z = lfbb.getCenterXYZ()
                    lfc = ncore.Point3D(x, y, z)
                    closestDistance = 1e32
                    for p in pins:
                        pbb = p.getBoundingBox()
                        px, py, pz = pbb.getCenterXYZ()
                        if not lfbb.is2DPointInBox([px, py]):
                            continue
                        pc = ncore.Point3D(px, py, pz)
                        dist = pc.distance(lfc)
                        if dist > closestDistance:
                            continue
                        pin = p
                        closestDistance = dist
                if not pin:
                    continue
                # print "Pin '{}' Found for Lead '{}'".format(pin, lf)
                oldName = pin.getName()
                newName = getNewPinName(pin, useCoordinates=False, names=names)
                # sp = lf.getStartPoint()
                # ep = lf.getStartPoint()
                pt = ncore.Point2D.getMiddlePoint(lf.getStartPoint(), lf.getEndPoint())
                x, y, z = pt.x(), pt.y(), lfbb.getMaxZ()
                newName += "_x%d_y%d_z%dnm" % (x * 1e9, y * 1e9, z * 1e9)
                newName = ensureUniqueName(newName, names)
                ro.addNames(oldName, newName)
                # pin.setName(newName)
                text += oldName + " = " + newName + "\n"
                renamedPins.append(oldName)
    # Rename the remaining Pins
    for pin in pins:
        oldName = pin.getName()
        if oldName in renamedPins:
            continue
        newName = getNewPinName(pin, useCoordinates, names)
        # pin.setName(newName)
        ro.addNames(oldName, newName)
        text += oldName + " = " + newName + "\n"
    model.rename(ro)
    # model.setDetails(md)
    if mapFile:
        with open(mapFile, 'wt') as fo:
            fo.write(text)
    # Call this to update the Pin Names in the Components:
    md = modelutils.getModelDetails(model)
    pins = md.getPins()
    for p in pins:
        # print "  {}".format(p.getName())
        p.setModifiedFlag(True)
    model.setDetails(md)
    pass


def writePortFile(fileName, model):
    """
    .. highlight:: xml

    Writes the Port Information to the given file from the given model in the format::

        <Ports>
          <Port name="<port name>" comp="<component name>" pin="<pin name>" net="<net name>" />
          <Port name="<port name>" comp="<component name>" pin="<pin name>" net="<net name>" />
        </Ports>

    Args:
      fileName (str): file name.
      model (ncore.Model): the Geometry Model.

    Example::

        <Ports>
          <Port name="1" comp="C01" pin="2" net="VOUT" />
          <Port name="2" comp="CINBULK" pin="1" net="VIN" />
          <Port name="3" comp="RSYNC" pin="1" net="SYNC" />
          <Port name="4" comp="U1" pin="14" net="VCC" />
        </Ports>

    """
    text = ""
    text += "<Ports>\n"
    po = model.getPortOptions()
    ports = po.getPorts()
    for port in ports:
        fullPinName = port.getSourcePinName()
        pinNameParts = fullPinName.split(".")
        comp, pin = "", ""
        if len(pinNameParts) > 0:
            comp = pinNameParts[0]
        if len(pinNameParts) > 1:
            pin = ".".join(pinNameParts[1:])
        text += "  <Port name=\"{}\" comp=\"{}\" pin=\"{}\" net=\"{}\" />\n".format(
            port.getName(), comp, pin, port.getSourceNetName())
    text += "</Ports>"
    with open(fileName, "wt") as pf:
        pf.write(text)
        pf.close()
    pass


def writePinDetailsFile(fileName, model, adjustLeadPins=True, leadSnapPoints=None):
    """
    .. highlight:: xml

    Writes the Pin Information to the given file from the given model in the format::

        <Pins>
          <Pin name="<name>" comp="<component>" pin="<pinOnly>" net="<net>" layer="<layer>"
            x="<xCenter(m)>" y="<yCenter(m)>" face="<placement>" />
        </Pins>

    Args:
      fileName (str): file name.
      model (ncore.Model): the Geometry Model.
      adjustLeadPins (bool): parameter to adjust the Pin coordinates for Lead Frame Pins
        (default True).
      leadSnapPoints (list of ncore.Point): list of points to snap to.

    Example::

        <Pins>
          <Pin name="DIE.A1" comp="DIE" pin="A1" net="GND" layer="TOP"
            x="0.123" y="0.456" placement="TOP" />
          <Pin name="LF1.PIN2" comp="LF1" pin="PIN2" net="NET1" layer="LF1"
            x="0.987" y="0.654" placement="BOTTOM" />
        </Pins>

    """
    text = ""
    text += "<Pins>\n"
    md = modelutils.getModelDetails(model)
    unit = md.getLengthUnit()
    md.setLengthUnit("m")
    pins = md.getPins()
    for pin in pins:
        fullPinName = pin.getName()
        pinNameOnly = pin.getPinName()
        comp = pin.getComponentName()
        net = pin.getNetName()
        layer = pin.getLayerName()
        bb = pin.getBoundingBox()
        x, y, z = bb.getCenterXYZ()
        if adjustLeadPins:
            # Adjust the Lead Pins to use the Midpoint of the connection:
            lfgs = md.getLeadFrameGroups()
            for lfg in lfgs:
                if pin.getLayerName() != lfg.getName():
                    continue
                # Find the nearest Lead Frame (Finger / Gull Wing):
                lfs = lfg.getLeadFrames()
                closestDistance = 1e32
                closestLeadFinger = None
                pbb = pin.getBoundingBox()
                px, py, pz = pbb.getCenterXYZ()
                pc = ncore.Point3D(px, py, pz)
                for lf in lfs:
                    # print lf
                    lfbb = lf.getBoundingBox()
                    x, y, z = lfbb.getCenterXYZ()
                    lfc = ncore.Point3D(x, y, z)
                    if not lfbb.is2DPointInBox([px, py]):
                        continue
                    dist = pc.distance(lfc)
                    if dist > closestDistance:
                        continue
                    closestDistance = dist
                    closestLeadFinger = lf
                if not closestLeadFinger:
                    continue
                lf = closestLeadFinger
                # print "Lead Finger '{}' Found for Pin '{}'".format(
                #     closestLeadFinger.getName(), pin.getName())
                # sp = lf.getStartPoint()
                # ep = lf.getStartPoint()
                pt = ncore.Point2D.getMiddlePoint(lf.getStartPoint(), lf.getEndPoint())
                if leadSnapPoints:
                    pt = getClosestPoint(leadSnapPoints, pt)
                lbb = lf.getBoundingBox()
                x, y, z = pt.x(), pt.y(), lbb.getMinZ()
        placement = pin.getPlacementString()
        text += "  <Pin name=\"{}\" comp=\"{}\" pin=\"{}\" net=\"{}\" layer=\"{}\"" \
                " x=\"{}\" y=\"{}\" placement=\"{}\" />\n" \
            .format(fullPinName, comp, pinNameOnly, net, layer, x, y, placement)
    text += "</Pins>"
    md.setLengthUnit(unit)
    with open(fileName, "wt") as pf:
        pf.write(text)
        pf.close()
    pass


def writeModelDetailsFile(fileName, model):
    """
    .. highlight:: xml

    Writes the Model Information to the given file from the given model in the format::

        <Model>
          <Nets>
            <Net name="<name>" />
          </Nets>
          <Nodes>
            <Node type="<type>" name="<name>" />
          </Nodes>
        </Model>

    Args:
      fileName (str): file name.
      model (ncore.Model): the Geometry Model.

    Example::

        <Model>
          <Nets>
            <Net name="NET001" />
            <Net name="NET002" />
          </Nets>
          <Nodes>
            <Node type="Pin" name="NET2_BGA-2" />
            <Node type="Pin" name="NET3_BGA-3" />
            <Node type="Net" name="NET2" />
            <Node type="Net" name="NET3" />
          </Nodes>
        </Model>

    """
    text = ""
    text += "<Model>\n"
    md = modelutils.getModelDetails(model)
    unit = md.getLengthUnit()
    md.setLengthUnit("m")
    nets = md.getNets()
    pins = md.getPins()
    text += "  <Nets>\n"
    for net in nets:
        name = net.getName()
        text += "    <Net name=\"{}\" />\n".format(name)
    text += "  </Nets>\n"
    text += "  <Nodes>\n"
    for pin in pins:
        name = "{}_{}-{}".format(pin.getNetName(), pin.getComponentName(), pin.getPinName())
        text += "    <Node type=\"Pin\" name=\"{}\" />\n".format(name)
    for net in nets:
        name = net.getName()
        text += "    <Node type=\"Net\" name=\"{}\" />\n".format(name)
    text += "  </Nodes>\n"
    text += "</Model>\n"
    md.setLengthUnit(unit)
    with open(fileName, "wt") as pf:
        pf.write(text)
        pf.close()
    pass


def getNetsAndComponentsInDesignFile(
        designFile, netsFile=None, techFile=None, individualPins=False):
    """
    Scan the Nets from the Design File and save the Net Name / Type to a File.

    Args:
      designFile (str): Design File or Project File Name.
      netsFile (str): File Name for Net Names, Types, and Components (e.g. "nets.list").
      techFile (str): File Name for Technology File (e.g. "technology.tech").
      individualPins (bool): Show the Component Names or Individual Pin names (default: False).
    """
    controller = ncore.Controller("nApex")
    ignoredNets = ["$NONE$", "<EMPTY_NET>"]
    project = None
    if designFile.lower().endswith(".phys"):
        project = controller.readProjectFile(designFile)
    else:
        project = controller.importDesign(designFile)
    model = project.getModel()
    if techFile:
        model.writeTechnologyFile(techFile)
    print("Getting Model Details...")
    md = model.getDetails()
    print("Getting Nets...")
    nets = md.getNets()
    netNamesAndTypes = []
    print("Net Names, Types, & Components:")
    for net in nets:
        if utils.nameMatches(net.getName(), ignoredNets):
            continue
        components = []
        pins = net.getPins()
        for pin in pins:
            cn = pin.getName() if individualPins else pin.getComponentName()
            if cn not in components:
                components.append(cn)
        nt = net.getNetType()  # Version 5.1.1
        try:
            nt = net.getNetTypeString()  # Version 6.0+
        except AttributeError:
            pass
        ntc = (net.getName(), nt, components)
        print("  %s" % str(ntc))
        netNamesAndTypes.append(ntc)
    if netsFile:
        with open(netsFile, "w") as fptr:
            fptr.write("%s" % netNamesAndTypes)
            fptr.close()
    pass


def exportSPICE(simulation, spiceFileName, spiceStages=5, modelType="PI", spiceSubstitutions=None,
                netlistNamePrefix=None, replacementSubckt=None, portMapFile=None,
                optionsFile=None):
    """
    Export the simulation SPICE file.

    Args:
      simulation (ncore.Simulation): the Simulation to export the SPICE model for.
      spiceFileName (str): the SPICE File Name to export to.
      spiceStages (int): Number of Stages to use in the SPICE Model [Default 5].
      modelType (str): "PI" or "L" [Default "PI"].
      spiceSubstitutions (dict): substitutions to use when exporting the SPICE model (default None)
      netlistNamePrefix (str): the name to use for the netlist Prefix.
      replacementSubckt (str): the name to use for the top circuit.
      portMapFile (str): the input Map file for renaming the Nodes.
      optionsFile (str): the Option File to use when writing the SPICE Model.

    Returns:
      str: full path name to the exported SPICE File.
    """
    # fn = super(SV_ApexFlow, self).exportSPICE(spiceFileName, modelType)
    c, p = None, None
    if utils.isType(simulation, ncore.Project):
        p = simulation
    elif utils.isType(simulation, str):
        c = ncore.Controller("nApex")
        p = c.readProjectFile(simulation)
    if p:
        simulations = p.getSimulations()
        if len(simulations) < 1:
            raise RuntimeError("Project '%s' has no Simulation" % p.getName())
        simulation = simulations[0]
    spiceFileName = os.path.abspath(spiceFileName)
    sfo = ncore.SpiceFileOptions()
    sfo.setFileName(spiceFileName)
    sfo.setModelType(modelType)  # "PI" or "L"
    sfo.setNumberOfStages(spiceStages)  # 1 to N
    sfo.setNetlistName("cMain")  # Any string
    sfo.setMaximumResistance(1e12)
    sfo.setIncludeMutual(True)  # True or False
    if optionsFile:
        sfo.setOptionFile(optionsFile)
    simulation.exportResults(sfo)
    if spiceSubstitutions or netlistNamePrefix or replacementSubckt or portMapFile:
        nnm = "%s.netnames.map" % spiceFileName
        convertSpiceToVirtuoso(spiceFileName, spiceFileName, spiceSubstitutions, nnm,
                               netlistNamePrefix=netlistNamePrefix,
                               replacementSubckt=replacementSubckt, portMapFile=portMapFile)
    return spiceFileName


def fixBoolFile(boolfile):
    with open(boolfile) as bf:
        lines = bf.readlines()
    if not lines:
        return
    with open(boolfile, "wt") as out:
        nets = False
        for line in lines:
            if line.startswith("BONDWIRE"):
                if nets:
                    # Remove the Net Information:
                    parts = line.split()
                    newline = " ".join([part for p, part in enumerate(parts) if p != 2])
                    out.write(newline)
                else:
                    out.write(line)
                out.write("\n")
            elif line.upper().startswith("NETS"):
                nets = line.split()[1].upper().startswith("Y")
                out.write("NETS NO\n")
            elif line.startswith("UNITS"):
                out.write(line)
        out.close()
    pass


def getShapePoints(shape):
    return [pt for pt in shape.getOutlinePoints()]


def getShapesPoints(shapes):
    points = list()
    for shape in shapes:
        points.extend(getShapePoints(shape))
    return points


def getLayerPoints(layer):
    return getShapesPoints(layer.getShapes())


def getLayersPoints(layers):
    points = list()
    for layer in layers:
        points.extend(getLayerPoints(layer))
    return points


def getClosestPoint(points, point):
    closestPoint = point
    closestDistance = 1e32
    for pt in points:
        # dist = point.distance(pt)
        dist = point.getDistanceTo(pt)
        if dist > closestDistance:
            continue
        closestDistance = dist
        closestPoint = pt
    return closestPoint



class MeshRefinement(object):
    @staticmethod
    def addNetFromPins(pins, netsUsedInPorts):
        for pin in pins:
            netName = pin.getNetName()
            if netName in netsUsedInPorts:
                continue
            netsUsedInPorts.add(netName)
        pass

    @staticmethod
    def addNetsFromPort(port, netsUsedInPorts):
        sourcePins = port.getSourcePins()
        sinkPins = port.getSinkPins()
        MeshRefinement.addNetFromPins(sourcePins, netsUsedInPorts)
        MeshRefinement.addNetFromPins(sinkPins, netsUsedInPorts)
        pass

    @staticmethod
    def addNetsFromPorts(ports, netsUsedInPorts):
        for port in ports:
            MeshRefinement.addNetsFromPort(port, netsUsedInPorts)
        pass

    @staticmethod
    def autoAdjustRefinement(simulation, refinementFactor):
        model = simulation.getModel()
        md = model.getDetails()
        netsUsedInPorts = set()
        ports = md.getPorts()
        circuitPorts = md.getCircuitPorts()
        shortedPorts = md.getShortedPorts()
        MeshRefinement.addNetsFromPorts(ports, netsUsedInPorts)
        MeshRefinement.addNetsFromPorts(circuitPorts, netsUsedInPorts)
        MeshRefinement.addNetsFromPorts(shortedPorts, netsUsedInPorts)
        # Loop over all Nets to find the smallest net diagonal
        nets = md.getNets()
        smallestNetDiagonal = 1e6
        modelDiag = md.getBoundingBox().getDiagonal()
        for net in nets:
            thisDiagonal = net.getBoundingBox().getDiagonal()
            if net.getName() not in netsUsedInPorts:
                continue
            if thisDiagonal < smallestNetDiagonal:
                smallestNetDiagonal = thisDiagonal
        mo = simulation.getMeshOptions()
        emo = mo.getExpertOptions()
        minimumRefinementFactor = refinementFactor
        if minimumRefinementFactor is None or minimumRefinementFactor < 0:
            if "MinimumRefinementFactor" in emo:
                minimumRefinementFactor = int(emo["MinimumRefinementFactor"])
            else:
                minimumRefinementFactor = 10
        smallestNetDiagonal = min(smallestNetDiagonal, modelDiag)
        criticalNetRefinementFactor = minimumRefinementFactor * int(modelDiag / smallestNetDiagonal)
        emo["CriticalNetRefinementFactor"] = str(criticalNetRefinementFactor)
        emo["MinimumRefinementFactor"] = str(minimumRefinementFactor)
        mo.setExpertOptions(emo)
        simulation.setMeshOptions(mo)
        pass
