-------------------------------------------------------------------------------
-- Copyright (c) 2010 Mentor Graphics Corporation
--
-- This model is a component of the SystemVision platform (www.mentor.com/sm)
-- and is covered by this license agreement. This model,
-- including any updates, modifications, revisions, copies, and documentation 
-- are copyrighted works of Mentor Graphics. USE OF THIS MODEL INDICATES YOUR 
-- COMPLETE AND UNCONDITIONAL ACCEPTANCE OF THE TERMS AND CONDITIONS SET FORTH
-- IN THIS LICENSE AGREEMENT.  Mentor Graphics grants you a non-exclusive 
-- license to use, reproduce, modify and distribute this model, provided that:
-- (a) no fee or other consideration is charged for any distribution except 
-- compilations distributed in accordance with Section (d) of this license 
-- agreement; (b) the comment text embedded in this model is included verbatim
-- in each copy of this model made or distributed by you, whether or not such 
-- version is modified; (c) any modified version must include a conspicuous 
-- notice that this model has been modified and the date of modification; and 
-- (d) any compilations sold by you that include this model must include a 
-- conspicuous notice that this model is available from Mentor Graphics in its
-- original form at no charge.
--
-- THIS MODEL IS LICENSED TO YOU "AS IS" AND WITH NO WARRANTIES, EXPRESS OR 
-- IMPLIED.  MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES OF 
-- MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  MENTOR GRAPHICS SHALL
-- HAVE NO RESPONSIBILITY FOR ANY DAMAGES WHATSOEVER.
-------------------------------------------------------------------------------
-- File       : capacitor_pcb_el_t.vhd
-- Author     : Mentor Graphics
-- Created    : 2010/05/04
-- Last update: 
-------------------------------------------------------------------------------
-- Description: Temperature-dependent capacitor with parasitics and power, energy calculations
-- This model consists of a capacitor in parallel with a leakage (current) resistance. 
-- This parallel C and R is connected to a series-connected ESR (equivalent-series-resistance), and 
-- ESL (equivalent-series-inductance) branch. Leakage current is a function of rated voltage.
-- ESR and ESL effects are optional, as is the leakage current. The capacitance is dependent on  
-- the global simulation temperature (set using .temp = <temp_value>. Default 
-- temperature is 27 degrees Celsius (default temperature units are Celsius).
-- This model also intenally calcualtes the capacitor's instantaneous power and 
-- energy. Initial voltage may be specified by typing a value for v_ic.
-- Also contains over-voltage and reverse-voltage detection
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version     Author              Description
-- 2010/05/04  1.0         Mentor Graphics     Created    
-------------------------------------------------------------------------------
library IEEE;
use IEEE.electrical_systems.all;
library MGC_AMS;
use MGC_AMS.ELDO_CONSTANTS.all; 	-- to get AmbientTemperatureCelsius value
use IEEE.energy_systems.all;		-- Required to recognize types "energy" and "power"

entity capacitor_pcb_el_t is

  generic (
    cap        : capacitance;         -- Nominal capacitance value [F]
    TC : real := 0.0;  				-- Capacitance temperature coefficient (first order only)
    temp_local : real := AmbientTempCelsius;  -- Temperature of this component instance, default to ambient
    v_rated    : voltage := 50.0;     -- Rated Maximum Voltage (must be greater than 0.0) [V]
    v_reverse_rated : voltage := 0.5; -- Rated maximum reverse voltage (must be greater than 0.0) [V]
    i_leak     : current := 0.0;      -- Leakage current at v_rated [Amps]
    esr        : resistance := 0.0;   -- Equivalent series resistance [Ohms]
    esl        : inductance := 0.0;   -- Equivalent series inductance [H]
    v_ic       : real := real'low;    -- Initial voltage [V]
    message_on : Boolean := False);   -- Transcript message reporting status [Default is False, no messages generated]
  port (
    terminal p1, p2 : electrical);

end entity capacitor_pcb_el_t;

-------------------------------------------------------------------------------
-- Ideal Architecture  (I = C * dV/dt)
-- Includes initial condition
-------------------------------------------------------------------------------
architecture level1 of capacitor_pcb_el_t is
  
  quantity v across i through p1 to p2;

begin

  if domain = quiescent_domain and v_ic /= real'low use
    v == v_ic;
  else
    i   == cap * v'dot;                  -- Fundamental equation
  end use;

end architecture level1;

-------------------------------------------------------------------------------
-- Advanced Architecture  Including ESR & ESL
-- Includes initial condition
-------------------------------------------------------------------------------
architecture level2 of capacitor_pcb_el_t is
  
  terminal mid : electrical; -- Internal terminal between parallel C and Gp (leakage conductance), and series esr and esL.

  quantity vc across ic, ip through p1 to mid;
  quantity v_esrl across i_esrl through mid to p2;
  quantity power_instantaneous  : power := 0.0;		-- Quantity to track pin-to-pin power
  quantity power_dissipated  : power := 0.0;		-- Quantity to track resistive power losses
  constant Gp : conductance := i_leak/v_rated;		-- Conductance parallel to capacitor
  constant temp_global : real := AmbientTempCelsius; -- "AmbientTemperatureCelsius" is the simulation global temperature 
                                                     -- expressed in Celsius, comes from package MGC_AMS.ELDO_CONSTANTS
  type electrolytic_state is (good, over_voltage, reverse_voltage);
  signal voltage_monitor : electrolytic_state := good;

begin
  -- Main capacitance effect, with optional initial voltage
  if domain = quiescent_domain and v_ic /= real'low use
    vc == v_ic;
  else
    ic == vc'dot*cap*(1.0 + TC*(temp_local - 27.0));	-- Capacitance as fucntion of temperature
  end use;

  -- Leakage (parallel) conductance effect
  ip == vc*Gp;
  
  -- Series resistance and inductance effects
  v_esrl == esr*i_esrl + esl*i_esrl'dot;

  -- Power calculations
  power_instantaneous  == i_esrl*(vc+v_esrl);	-- Pin-to-pin power
  power_dissipated == vc*ip + v_esrl*i_esrl;	-- Power dissipated by leakage conductance and ESR

  -- Over Voltage and Reverse Voltage Stress Detection Process
  voltage_stress_detection : process (vc'ABOVE(v_rated), vc'above(-1.0*v_reverse_rated))  
  begin
    if vc'ABOVE(v_rated) then
      voltage_monitor <= over_voltage;
      if message_on then
        report " Over Voltage detected at time = " & real'IMAGE(NOW);
      end if;
    elsif not vc'ABOVE(-1.0*v_reverse_rated) then
      voltage_monitor <= reverse_voltage;
      if message_on then
        report " Reverse Voltage detected at time = " & real'IMAGE(NOW);
      end if; 
    else
      voltage_monitor <= good;
    end if;
  end process voltage_stress_detection;
  
end architecture level2;

-------------------------------------------------------------------------------
-- Copyright (c) 2010 Mentor Graphics Corporation
-------------------------------------------------------------------------------
