-------------------------------------------------------------------------------
-- Copyright (c) 2017 Mentor Graphics Corporation
--
-- This model is a component of the SystemVision platform (www.mentor.com/sm)
-- and is covered by this license agreement. This model,
-- including any updates, modifications, revisions, copies, and documentation 
-- are copyrighted works of Mentor Graphics. USE OF THIS MODEL INDICATES YOUR 
-- COMPLETE AND UNCONDITIONAL ACCEPTANCE OF THE TERMS AND CONDITIONS SET FORTH
-- IN THIS LICENSE AGREEMENT.  Mentor Graphics grants you a non-exclusive 
-- license to use, reproduce, modify and distribute this model, provided that:
-- (a) no fee or other consideration is charged for any distribution except 
-- compilations distributed in accordance with Section (d) of this license 
-- agreement; (b) the comment text embedded in this model is included verbatim
-- in each copy of this model made or distributed by you, whether or not such 
-- version is modified; (c) any modified version must include a conspicuous 
-- notice that this model has been modified and the date of modification; and 
-- (d) any compilations sold by you that include this model must include a 
-- conspicuous notice that this model is available from Mentor Graphics in its
-- original form at no charge.
--
-- THIS MODEL IS LICENSED TO YOU "AS IS" AND WITH NO WARRANTIES, EXPRESS OR 
-- IMPLIED.  MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES OF 
-- MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  MENTOR GRAPHICS SHALL
-- HAVE NO RESPONSIBILITY FOR ANY DAMAGES WHATSOEVER.
-------------------------------------------------------------------------------
-- File       : resistor_pcb.vhd
-- Author     : Mentor Graphics
-- Created    : 2010/01/06
-- Last update: 2017/04/21
-------------------------------------------------------------------------------
-- Description: Temperature-dependent resistor with ESL and power, energy calculations
-- This is a simple resistor model with ESL (equivalent-series inductance).
-- ESL is optional. The resistance of this resistor is dependent on 
-- the global simulation temperature (set using .temp = <temp_value>. Default 
-- temperature is 27 degrees Celsius (default temperature units are Celsius).
-- This model also internally calculates the resistor's instantaneous power and 
-- energy.
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version     Author              Description
-- 2010/01/06  1.0         Mentor Graphics     Created 
-- 2017/04/21  1.1         Mentor Graphics     Added local temperature generic
-------------------------------------------------------------------------------

library IEEE;
use IEEE.electrical_systems.all;
library MGC_AMS;
use MGC_AMS.ELDO_CONSTANTS.all; 	-- to get AmbientTemperatureCelsius value
use IEEE.energy_systems.all;		-- Required to recognize types "energy" and "power"

entity resistor_pcb_t is
  generic (
    res : resistance;               -- User must enter nominal resistance value at 27 degrees C [Ohms] 
    esl : inductance := 0.0;		-- Equivalent Series Inductance (H]
    TC : real := 0.0;  			-- Resistance temperature coefficient (first order only) 
    temp_local : real := AmbientTempCelsius;  -- Temperature of this component instance, default to ambient
    power_peak_max : power := 1.0E6; -- Maximum allowed peak (instantaneous) power [Watts]
    message_on : Boolean := False);   -- Transcript message reporting status [Default is False, no messages generated]
  port (
    terminal p1, p2 : electrical);

end entity resistor_pcb_t;

-------------------------------------------------------------------------------
-- Level1 Architecture (V = I*R)
-------------------------------------------------------------------------------
architecture level1 of resistor_pcb_t is

  quantity v across i through p1 to p2;

begin

-- Fundamental equation
  v == i*res;

end architecture level1;


-------------------------------------------------------------------------------
-- Level2 Architecture ads power, esl, TC
-------------------------------------------------------------------------------
architecture level2 of resistor_pcb_t is

  quantity v across i through p1 to p2;
  quantity power_instantaneous  : power := 0.0;		-- Quantity to track pin-to-pin power
  quantity power_dissipated  : power := 0.0;		-- Quantity to track resistive power losses
  constant temp_global : real := AmbientTempCelsius; -- "AmbientTemperatureCelsius" is the simulation global temperature 
                                                     -- expressed in Celsius, comes from package MGC_AMS.ELDO_CONSTANTS
  type res_stress_state is (good, over_power);
  signal stress_monitor : res_stress_state := good;

begin

  v == i*res*(1.0 + TC*(temp_local - 27.0)) + esl*i'dot;		-- Resistance as fucntion of temperature

  -- Power calculations
  power_instantaneous  == v*i;	-- Pin-to-pin power
  power_dissipated == res*i**2;	-- Power dissipated by resistance

  -- Over Power Stress Detection Process
  power_stress_detection : process (power_instantaneous'ABOVE(power_peak_max))  
  begin
    if power_instantaneous'ABOVE(power_peak_max) then
      stress_monitor <= over_power;
      if message_on then
        report " Over Power detected at time = " & real'IMAGE(NOW);
      end if;
    else
      stress_monitor <= good;
    end if;
  end process power_stress_detection;

end architecture level2;

-------------------------------------------------------------------------------
-- Copyright (c) 2017 Mentor Graphics Corporation
-------------------------------------------------------------------------------
